import json
import os
import time
import traceback

from studio.artifacts.artifact import Artifact
from studio.db_providers import db_provider_setup
from studio.experiments.experiment import Experiment
from studio import git_util
from studio.payload_builders.payload_builder import PayloadBuilder
from studio.payload_builders.unencrypted_payload_builder import UnencryptedPayloadBuilder
from studio.encrypted_payload_builder import EncryptedPayloadBuilder
from studio.storage import storage_setup
from studio.util import util

def submit_experiments(
        experiments,
        config=None,
        logger=None,
        queue=None,
        python_pkg=[],
        external_payload_builder: PayloadBuilder=None):

    num_experiments = len(experiments)

    payload_builder = external_payload_builder
    if payload_builder is None:
        # Setup our own payload builder
        payload_builder = UnencryptedPayloadBuilder("simple-payload")
        # Are we using experiment payload encryption?
        public_key_path = config.get('public_key_path', None)
        if public_key_path is not None:
            logger.info("Using RSA public key path: {0}".format(public_key_path))
            signing_key_path = config.get('signing_key_path', None)
            if signing_key_path is not None:
                logger.info("Using RSA signing key path: {0}".format(signing_key_path))
            payload_builder = \
                EncryptedPayloadBuilder(
                    "cs-rsa-encryptor [{0}]".format(public_key_path),
                    public_key_path, signing_key_path)

    start_time = time.time()

    # Reset our storage setup, which will guarantee
    # that we rebuild our database and storage provider objects
    # that's important in the case that previous experiment batch
    # cleaned up after itself.
    storage_setup.reset_storage()

    for experiment in experiments:
        # Update Python environment info for our experiments:
        experiment.pythonenv = util.add_packages(experiment.pythonenv, python_pkg)

        # Add experiment to database:
        try:
            with db_provider_setup.get_db_provider(config) as db:
                _add_git_info(experiment, logger)
                db.add_experiment(experiment)
        except BaseException:
            traceback.print_exc()
            raise

        payload = payload_builder.construct(experiment, config, python_pkg)

        logger.info("Submitting experiment: {0}"
                     .format(json.dumps(payload, indent=4)))

        queue.enqueue(json.dumps(payload))
        logger.info("studio run: submitted experiment " + experiment.key)

    logger.info("Added {0} experiment(s) in {1} seconds to queue {2}"
                .format(num_experiments, int(time.time() - start_time), queue.get_name()))
    return queue.get_name()

def _add_git_info(experiment: Experiment, logger):
    wrk_space: Artifact = experiment.artifacts.get('workspace', None)
    if wrk_space is not None:
        if wrk_space.local_path is not None and \
                os.path.exists(wrk_space.local_path):
            if logger is not None:
                logger.info("git location for experiment %s", wrk_space.local_path)
            experiment.git = git_util.get_git_info(wrk_space.local_path)



try:
    import boto3
    import botocore
except BaseException:
    boto3 = None
    botocore = None

import uuid
import os
import base64
import json
import yaml
import six
import filelock
import hashlib
import pickle

from studio import git_util
from studio.util import logs
from studio.storage import storage_util
from studio.util.gpu_util import memstr2int
from .cloud_worker_util import insert_user_startup_script

# list of instance types sorted by price
_instance_specs = {
    'c4.large': {
        'cpus': 2,
        'ram': '3.75g',
        'gpus': 0
    },

    'c4.xlarge': {
        'cpus': 4,
        'ram': '7.5g',
        'gpus': 0
    },
    'c4.2xlarge': {
        'cpus': 8,
        'ram': '15g',
        'gpus': 0
    },
    'c4.4xlarge': {
        'cpus': 16,
        'ram': '30g',
        'gpus': 0
    },
    'p2.xlarge': {
        'cpus': 4,
        'ram': '61g',
        'gpus': 1
    },
    'c4.8xlarge': {
        'cpus': 36,
        'ram': '60g',
        'gpus': 0
    },
    'p2.8xlarge': {
        'cpus': 32,
        'ram': '488g',
        'gpus': 8
    },
    'p2.16xlarge': {
        'cpus': 64,
        'ram': '732g',
        'gpus': 16
    }
}


class EC2WorkerManager(object):

    def __init__(self, auth_cookie=None, verbose=10, branch=None,
                 user_startup_script=None):
        self.startup_script_file = os.path.join(
            os.path.dirname(__file__),
            'scripts/ec2_worker_startup.sh')

        self.install_studio_script = os.path.join(
            os.path.dirname(__file__),
            'scripts/install_studio.sh')

        self.client = boto3.client('ec2')
        self.asclient = boto3.client('autoscaling')
        self.cwclient = boto3.client('cloudwatch')

        self.region = self.client._client_config.region_name

        self.logger = logs.getLogger('EC2WorkerManager')
        self.logger.setLevel(verbose)
        self.auth_cookie = auth_cookie

        self.prices = self._get_ondemand_prices(_instance_specs.keys())

        self.repo_url = git_util.get_my_repo_url()
        self.branch = branch if branch else git_util.get_my_checkout_target()
        self.user_startup_script = user_startup_script

        if user_startup_script:
            self.logger.warn('User startup script argument is deprecated')

    def _get_image_id(self):
        price_path = os.path.join(
            os.path.dirname(__file__),
            'aws/aws_amis.yaml')
        with open(price_path) as f:
            ami_dict = yaml.load(f.read(), Loader=yaml.FullLoader)

        region = self.client._client_config.region_name
        image_type = 'ubuntu16.04'
        return ami_dict[image_type][region]

    def _get_block_device_mappings(self, resources_needed):
        return [{
            'DeviceName': '/dev/sda1',
            'Ebs': {
                'DeleteOnTermination': True,
                'VolumeSize': int(memstr2int(resources_needed['hdd']) /
                                  memstr2int('1g')),
                'VolumeType': 'standard'
            }
        }]

    def start_worker(
            self,
            queue_name,
            resources_needed={},
            blocking=True,
            ssh_keypair=None,
            timeout=300,
            ports=[]):

        imageid = self._get_image_id()

        name = self._generate_instance_name()

        instance_type = self._select_instance_type(resources_needed)

        startup_script = self._get_startup_script(
            resources_needed, queue_name, timeout=timeout)

        self.logger.info(
            'Starting EC2 instance of type {}'.format(instance_type))
        kwargs = {
            'BlockDeviceMappings':
                self._get_block_device_mappings(resources_needed),
            'ImageId': imageid,
            'InstanceType': instance_type,
            'MaxCount': 1,
            'MinCount': 1,
            'UserData': startup_script,
            'InstanceInitiatedShutdownBehavior': 'terminate',
            'TagSpecifications': [{
                'ResourceType': 'instance',
                'Tags': [
                    {
                        'Key': 'Name',
                        'Value': name
                    },
                    {
                        'Key': 'Team',
                        'Value': 'tools'
                    }]
            }]
        }
        ports = set(ports)
        if ssh_keypair is not None:
            ports.add(22)  # ssh port
        if any(ports):
            groupid = self._get_security_group(ports)
            kwargs['SecurityGroupIds'] = [groupid]
            if ssh_keypair is not None:
                kwargs['KeyName'] = ssh_keypair

        response = self.client.run_instances(**kwargs)
        instance_id = response['Instances'][0]['InstanceId']
        self.logger.info(
            'Starting instance {}'.format(instance_id))

        if blocking:
            while True:
                try:
                    response = self.client.describe_instances(
                        InstanceIds=[instance_id]
                    )
                    instance_data = response['Reservations'][0]['Instances'][0]
                    ip_addr = instance_data.get('PublicIpAddress')
                    if ip_addr:
                        print("ip address: {}".format(ip_addr))
                        return
                except BaseException as e:
                    pass

    def _select_instance_type(self, resources_needed):
        sorted_specs = sorted(_instance_specs.items(),
                              key=lambda x: self.prices[x[0]])
        for instance in sorted_specs:
            if int(
                instance[1]['cpus']) >= int(
                resources_needed['cpus']) and memstr2int(
                instance[1]['ram']) >= memstr2int(
                resources_needed['ram']) and int(
                    instance[1]['gpus']) >= int(
                        resources_needed['gpus']):
                return instance[0]

        raise ValueError('No instances that satisfy requirements {} '
                         'can be found'.format(resources_needed))

    def _get_startup_script(
            self,
            resources_needed,
            queue_name,
            autoscaling_group=None,
            timeout=300):
        if self.auth_cookie is not None:
            auth_key = os.path.basename(self.auth_cookie)
            with open(self.auth_cookie, 'r') as f:
                auth_data = f.read()
        else:
            auth_key = None
            auth_data = None

        credentials = ''

        if 'GOOGLE_APPLICATION_CREDENTIALS' in os.environ.keys():
            with open(os.environ['GOOGLE_APPLICATION_CREDENTIALS'], 'r') as f:
                credentials = f.read()
        else:
            self.logger.info('credentials NOT found')

        with open(self.startup_script_file) as f:
            startup_script = f.read()

        with open(self.install_studio_script) as f:
            install_studio_script = f.read()

        startup_script = startup_script.replace(
            '{install_studio}', install_studio_script)

        startup_script = startup_script.format(
            auth_key=auth_key if auth_key else "",
            queue_name=queue_name,
            auth_data=base64.b64encode(auth_data.encode('utf-8'))
            .decode('utf-8') if auth_data else "",
            google_app_credentials=base64.b64encode(
                credentials.encode('utf-8')).decode('utf-8'),
            aws_access_key=self.client._request_signer._credentials.access_key,
            aws_secret_key=self.client._request_signer._credentials.secret_key,
            autoscaling_group=autoscaling_group if autoscaling_group else "",
            region=self.region,
            use_gpus=0 if resources_needed['gpus'] == 0 else 1,
            timeout=timeout,
            repo_url=self.repo_url,
            studioml_branch=self.branch,
        )

        startup_script = insert_user_startup_script(
            self.user_startup_script,
            startup_script, self.logger)

        self.logger.info('Startup script:')
        self.logger.info(startup_script)

        return startup_script

    def _generate_instance_name(self):
        return 'studioml_worker_' + str(uuid.uuid4())

    def _get_security_group(self, ports):
        ports = sorted([p for p in set(ports)])
        group_name = "studioml_{}".format(
            hashlib.sha256(
                pickle.dumps(sorted(ports))
            ).hexdigest()
        )

        try:
            response = self.client.describe_security_groups(
                GroupNames=[group_name]
            )
            groupid = response['SecurityGroups'][0]['GroupId']

        except botocore.exceptions.ClientError as e:

            self.logger.error("Error creating security group!")
            self.logger.exception(e)
            response = self.client.create_security_group(
                GroupName=group_name,
                Description='opens ports {} in studioml workers'
                            .format(",".join([str(p) for p in ports]))
            )

            groupid = response['GroupId']

            ip_permissions = []
            for port in ports:
                ip_permissions.append({
                    'IpProtocol': 'tcp',
                    'FromPort': int(port),
                    'ToPort': int(port),
                    'IpRanges': [{
                        'CidrIp': '0.0.0.0/0'
                    }]
                })

            response = self.client.authorize_security_group_ingress(
                GroupId=groupid,
                GroupName=group_name,
                IpPermissions=ip_permissions
            )

        return groupid

    def start_spot_workers(
            self,
            queue_name,
            bid_price,
            resources_needed={},
            ssh_keypair=None,
            queue_upscaling=True,
            start_workers=1,
            max_workers=100,
            timeout=300,
            ports=[],
            autoscaling_group_name=None):

        # TODO should be able to put bid price as None,
        # which means price of on-demand instance
        # or maybe specify bid_price as a fraction of on-demand price

        instance_type = self._select_instance_type(resources_needed)

        asg_name = "studioml_autoscaling_" + queue_name or \
                   autoscaling_group_name
        launch_config_name = asg_name + "_launch_config"

        startup_script = self._get_startup_script(
            resources_needed, queue_name, asg_name, timeout=timeout)

        if bid_price.endswith('%'):
            bid_price = str(self.prices[instance_type] *
                            float(bid_price.replace('%', '')) / 100)

        self.logger.info('Price bid for instance type {} : {}'
                         .format(instance_type, bid_price))

        launch_config = {
            "ImageId": self._get_image_id(),
            "UserData": startup_script,
            "InstanceType": instance_type,
            "BlockDeviceMappings": self._get_block_device_mappings(
                resources_needed),
            "InstanceMonitoring": {
                'Enabled': False},
            "SpotPrice": bid_price,
        }

        launch_config_name = 'studioml_launch_config_' + \
            hashlib.sha256(
                json.dumps(launch_config, sort_keys=True).encode('utf-8')
            ).hexdigest()

        if ssh_keypair is not None:
            ports.append(22)
        if any(ports):
            groupid = self._get_security_group(ports)
            launch_config['SecurityGroups'] = [groupid]
            if ssh_keypair is not None:
                launch_config['KeyName'] = ssh_keypair

        try:
            response = self.asclient.create_launch_configuration(
                LaunchConfigurationName=launch_config_name, **launch_config)
            self.logger.debug(
                "create_launch_configuration response:\n {}".format(response))
        except self.asclient.exceptions.AlreadyExistsFault:
            self.logger.debug('Launch config {} already exists'
                              .format(launch_config_name))

        asg_config = {
            "LaunchConfigurationName": launch_config_name,
            "MinSize": 0,
            "MaxSize": max_workers,
            "DesiredCapacity": int(start_workers),
            "LoadBalancerNames": [],
            "AvailabilityZones": [self.region + "a"],
            "TerminationPolicies": ['NewestInstance'],
            "DefaultCooldown": 0,
            "NewInstancesProtectedFromScaleIn": True
        }

        self.logger.debug("Creating auto-scaling group " + asg_name)

        try:
            response = self.asclient.create_auto_scaling_group(
                AutoScalingGroupName=asg_name, **asg_config)
        except self.asclient.exceptions.AlreadyExistsFault:
            self.logger.debug('Autoscaling group {} already exists'
                              .format(asg_name))

        if queue_upscaling:
            max_steps = 20
            stepsize = - (- max_workers // max_steps)

            step_adjustments = [{
                'MetricIntervalLowerBound': i,
                'MetricIntervalUpperBound': i + stepsize,
                'ScalingAdjustment': i
            } for i in range(0, max_workers, stepsize)]

            del step_adjustments[-1]['MetricIntervalUpperBound']

            scaleup_policy_response = self.asclient.put_scaling_policy(
                AutoScalingGroupName=asg_name,
                PolicyName=asg_name + "_scaleup",
                PolicyType='StepScaling',
                AdjustmentType="ChangeInCapacity",
                StepAdjustments=step_adjustments
            )

            self.cwclient.put_metric_alarm(
                AlarmName=asg_name + "_scaleup_alarm",
                MetricName="ApproximateNumberOfMessagesVisible",
                Namespace="AWS/SQS",
                Statistic="Average",
                Period=300,
                Threshold=1,
                ComparisonOperator="GreaterThanOrEqualToThreshold",
                EvaluationPeriods=1,
                AlarmActions=[scaleup_policy_response['PolicyARN']],
                Dimensions=[{
                    'Name': 'QueueName',
                    'Value': queue_name
                }]
            )

    def _get_ondemand_prices(self, instances=_instance_specs.keys()):

        # TODO un-hardcode the us-east as a region
        # so that prices are being read for a correct region

        price_path = os.path.join(
            os.path.dirname(__file__),
            'aws/aws_prices.yaml')
        with open(price_path, 'r') as f:
            data = yaml.load(f.read(), Loader=yaml.FullLoader)

        return {i: data[i] for i in instances}

        price_path = os.path.join(os.path.expanduser('~'), '.studioml',
                                  'awsprices.json')
        offer_file_lock = filelock.FileLock(price_path + '.lock')

        try:
            self.logger.info('Reading AWS prices from cache...')
            with offer_file_lock:
                with open(price_path, 'r') as f:
                    offer_dict = json.load(f)

        except IOError:
            self.logger.info(
                'Getting prices info from AWS (this may take a moment...)')

            with offer_file_lock:
                storage_util.download_file(
                    'https://pricing.us-east-1.amazonaws.com/offers/v1.0/'
                    'aws/AmazonEC2/current/index.json',
                    price_path)

                with open(price_path, 'r') as f:
                    offer_dict = json.load(f)

        self.logger.info('Done!')

        region_name = 'US East (N. Virginia)'

        prices = {}

        for instance_type in instances:
            product_sku = [
                k for k, v in six.iteritems(offer_dict['products'])
                if v['attributes'].get('instanceType') == instance_type and
                v['attributes']['tenancy'] == 'Shared' and
                v['attributes']['operatingSystem'] == 'Linux' and
                v['attributes']['location'] == region_name
            ]

            assert len(product_sku) == 1, \
                'Either no or too many products found for {}!' \
                .format(instance_type)

            prices[instance_type] = float(
                list(
                    list(
                        offer_dict['terms']['OnDemand']
                        [product_sku[0]].values()
                    )[0]['priceDimensions'].values()
                )[0]['pricePerUnit']['USD'])

        return prices



import os
import json

from studio.db_providers.keyvalue_provider import KeyValueProvider
from studio.storage.storage_handler_factory import StorageHandlerFactory
from studio.storage.storage_type import StorageType
from studio.util import util

class LocalDbProvider(KeyValueProvider):

    def __init__(self, config, blocking_auth=True):
        self.config = config
        self.bucket = config.get('bucket', 'studioml-meta')

        factory: StorageHandlerFactory = StorageHandlerFactory.get_factory()
        self.meta_store = factory.get_handler(StorageType.storageLocal, config)

        self.endpoint = self.meta_store.get_endpoint()
        self.db_root = os.path.join(self.endpoint, self.bucket)
        self._ensure_path_dirs_exist(self.db_root)

        super().__init__(config, self.meta_store, blocking_auth)

    def _ensure_path_dirs_exist(self, path):
        dirs = os.path.dirname(path)
        os.makedirs(dirs, mode = 0o777, exist_ok = True)

    def _get(self, key, shallow=False):
        file_name = os.path.join(self.db_root, key)
        if not os.path.exists(file_name):
            return None
        with open(file_name) as infile:
            result = json.load(infile)
        return result

    def _delete(self, key, shallow=True):
        file_name = os.path.join(self.db_root, key)
        if os.path.exists(file_name):
            self.logger.debug("Deleting local database file %s.", file_name)
            util.delete_local_path(file_name, self.db_root, shallow)

    def _set(self, key, value):
        file_name = os.path.join(self.db_root, key)
        self._ensure_path_dirs_exist(file_name)
        with open(file_name, 'w') as outfile:
            json.dump(value, outfile)



import json
from studio.db_providers.keyvalue_provider import KeyValueProvider
from studio.storage.storage_handler_factory import StorageHandlerFactory
from studio.storage.storage_type import StorageType


class S3Provider(KeyValueProvider):

    def __init__(self, config, blocking_auth=True):
        self.config = config
        self.bucket = config.get('bucket', 'studioml-meta')

        factory: StorageHandlerFactory = StorageHandlerFactory.get_factory()
        self.meta_store = factory.get_handler(StorageType.storageS3, config)

        super().__init__(
            config,
            self.meta_store,
            blocking_auth)

    def _get(self, key, shallow=False):
        try:
            response = self.meta_store.client.list_objects(
                Bucket=self.bucket,
                Prefix=key,
                Delimiter='/',
                MaxKeys=1024*16
            )
        except Exception as exc:
            msg: str = "FAILED to list objects in bucket {0}: {1}"\
                .format(self.bucket, exc)
            self._report_fatal(msg)
            return None

        if response is None:
            return None

        if 'Contents' not in response.keys():
            return None

        key_count = len(response['Contents'])

        if key_count == 0:
            return None

        for key_item in response['Contents']:
            if 'Key' in key_item.keys() and key_item['Key'] == key:
                response = self.meta_store.client.get_object(
                    Bucket=self.bucket,
                    Key=key)
                return json.loads(response['Body'].read().decode("utf-8"))

        return None

    def _delete(self, key, shallow=True):
        self.logger.info("S3 deleting object: %s/%s", self.bucket, key)

        try:
            response = self.meta_store.client.delete_object(
                Bucket=self.bucket,
                Key=key)
        except Exception as exc:
            msg: str = "FAILED to delete object {0} in bucket {1}: {2}"\
                .format(key, self.bucket, exc)
            self.logger.info(msg)
            return

        reason = response['ResponseMetadata'] if response else "None"
        if response is None or\
            response['ResponseMetadata']['HTTPStatusCode'] != 204:
            msg: str = ('attempt to delete key {0} in bucket {1}' +
                       ' returned response {2}')\
                           .format(key, self.bucket, reason)
            self.logger.info(msg)

    def _set(self, key, value):
        try:
            response = self.meta_store.client.put_object(
                Bucket=self.bucket,
                Key=key,
                Body=json.dumps(value))
        except Exception as exc:
            msg: str = "FAILED to write object {0} in bucket {1}: {2}"\
                .format(key, self.bucket, exc)
            self._report_fatal(msg)

        reason = response['ResponseMetadata'] if response else "None"
        if response is None or \
                response['ResponseMetadata']['HTTPStatusCode'] != 200:
            msg: str = ('attempt to write key {0} in bucket {1}' +
                       ' returned response {2}')\
                .format(key, self.bucket, reason)
            self._report_fatal(msg)



"""Data providers."""
import os
import yaml
import pyhocon

from studio.db_providers.local_db_provider import LocalDbProvider
from studio.db_providers.s3_provider import S3Provider
from studio.storage.storage_handler import StorageHandler
from studio.storage.storage_handler_factory import StorageHandlerFactory
from studio.storage.storage_setup import setup_storage, get_storage_db_provider,\
    set_storage_verbose_level
from studio.storage.storage_type import StorageType
from studio.util import logs
from studio.util.util import parse_verbosity

def get_config(config_file=None):

    config_paths = []
    if config_file:
        if not os.path.exists(config_file):
            raise ValueError('User config file {} not found'
                             .format(config_file))
        config_paths.append(os.path.expanduser(config_file))

    config_paths.append(os.path.expanduser('~/.studioml/config.yaml'))
    config_paths.append(
        os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "default_config.yaml"))

    for path in config_paths:
        if not os.path.exists(path):
            continue

        with(open(path)) as f_in:
            if path.endswith('.hocon'):
                config = pyhocon.ConfigFactory.parse_string(f_in.read())
            else:
                config = yaml.load(f_in.read(), Loader=yaml.FullLoader)

                def replace_with_env(config):
                    for key, value in config.items():
                        if isinstance(value, str):
                            config[key] = os.path.expandvars(value)

                        elif isinstance(value, dict):
                            replace_with_env(value)

                replace_with_env(config)

            return config

    raise ValueError('None of the config paths {0} exists!'
                     .format(config_paths))

def get_artifact_store(config) -> StorageHandler:
    storage_type: str = config['type'].lower()

    factory: StorageHandlerFactory = StorageHandlerFactory.get_factory()
    if storage_type == 's3':
        handler = factory.get_handler(StorageType.storageS3, config)
        return handler
    if storage_type == 'local':
        handler = factory.get_handler(StorageType.storageLocal, config)
        return handler
    raise ValueError('Unknown storage type: ' + storage_type)

def get_db_provider(config=None, blocking_auth=True):

    db_provider = get_storage_db_provider()
    if db_provider is not None:
        return db_provider

    if config is None:
        config = get_config()
    verbose = parse_verbosity(config.get('verbose'))

    # Save this verbosity level as global for the whole experiment job:
    set_storage_verbose_level(verbose)

    logger = logs.getLogger("get_db_provider")
    logger.setLevel(verbose)
    logger.debug('Choosing db provider with config:')
    logger.debug(config)

    if 'storage' in config.keys():
        artifact_store = get_artifact_store(config['storage'])
    else:
        artifact_store = None

    assert 'database' in config.keys()
    db_config = config['database']
    if db_config['type'].lower() == 's3':
        db_provider = S3Provider(db_config,
                          blocking_auth=blocking_auth)
        if artifact_store is None:
            artifact_store = db_provider.get_storage_handler()

    elif db_config['type'].lower() == 'gs':
        raise NotImplementedError("GS is not supported.")

    elif db_config['type'].lower() == 'local':
        db_provider = LocalDbProvider(db_config,
                          blocking_auth=blocking_auth)
        if artifact_store is None:
            artifact_store = db_provider.get_storage_handler()

    else:
        raise ValueError('Unknown type of the database ' + db_config['type'])

    setup_storage(db_provider, artifact_store)
    return db_provider



"""Data providers."""
import uuid

from studio.queues.local_queue import LocalQueue
from studio.queues.sqs_queue import SQSQueue
from studio.queues.qclient_cache import get_cached_queue, shutdown_cached_queue
from studio.util.util import parse_verbosity

from studio.db_providers import db_provider_setup
from studio.storage.storage_setup import setup_storage, get_storage_db_provider,\
    reset_storage, set_storage_verbose_level
from studio.util import logs

def get_queue(
        queue_name=None,
        cloud=None,
        config=None,
        logger=None,
        close_after=None,
        verbose=10):
    if queue_name is None:
        if cloud in ['gcloud', 'gcspot']:
            queue_name = 'pubsub_' + str(uuid.uuid4())
        elif cloud in ['ec2', 'ec2spot']:
            queue_name = 'sqs_' + str(uuid.uuid4())
        else:
            queue_name = 'local'

    if queue_name.startswith('ec2') or \
       queue_name.startswith('sqs'):
        return SQSQueue(queue_name, config=config, logger=logger)
    elif queue_name.startswith('rmq_'):
        return get_cached_queue(
            name=queue_name,
            route='StudioML.' + queue_name,
            config=config,
            close_after=close_after,
            logger=logger,
            verbose=verbose)
    elif queue_name == 'local':
        return LocalQueue(verbose=verbose)
    else:
        return None

def shutdown_queue(queue, logger=None, delete_queue=True):
    if queue is None:
        return
    queue_name = queue.get_name()
    if queue_name.startswith("rmq_"):
        shutdown_cached_queue(queue, logger, delete_queue)
    else:
       queue.shutdown(delete_queue)



import openssh_key.private_key_list as pkl
from openssh_key.key import PublicKey

class Ed25519KeyUtil:

    @classmethod
    def parse_private_key_file(cls, filepath: str, logger):
        """
        Parse a file with ed25519 private key in OPEN SSH PRIVATE key format.
        :param filepath: file path to a key file
        :param logger: logger to use for messages
        :return: (public key part as bytes, private key part as bytes)
        """
        contents: str = ''
        try:
            with open(filepath, "r") as f:
                contents = f.read()
        except Exception as exc:
            if logger is not None:
                logger.error("FAILED to read keyfile %s: %s",
                             filepath, exc)
            return None, None
        try:
            key_data = pkl.PrivateKeyList.from_string(contents)
            data_public = key_data[0].private.params['public']
            data_private = key_data[0].private.params['private_public']
            return data_public, data_private[:32]
        except Exception as exc:
            if logger is not None:
                logger.error("FAILED to decode keyfile format %s: %s",
                         filepath, exc)
            return None, None

    @classmethod
    def parse_public_key_file(cls, filepath: str, logger):
        """
        Parse a file with ed25519 public.
        :param filepath: file path to a key file
        :param logger: logger to use for messages
        :return: public key part as bytes
        """
        contents: str = ''
        try:
            with open(filepath, "r") as f:
                contents = f.read()
        except Exception as exc:
            if logger is not None:
                logger.error("FAILED to read keyfile %s: %s",
                             filepath, exc)
            return None, None
        try:
            key_data = PublicKey.from_string(contents)
            data_public = key_data.params['public']
            return data_public
        except Exception as exc:
            if logger is not None:
                logger.error("FAILED to decode keyfile format %s: %s",
                         filepath, exc)
            return None



import re

try:
    try:
        from pip._internal.operations import freeze
    except Exception:
        from pip.operations import freeze
except ImportError:
    freeze = None

from studio.dependencies_policies.dependencies_policy import DependencyPolicy

class StudioDependencyPolicy(DependencyPolicy):
    """
    StudioML policy for adjusting experiment dependencies
    to specific execution environment and required resources.
    """

    def __init__(self):
        super(DependencyPolicy, self).__init__()

    def generate(self, resources_needed):
        if freeze is None:
            raise ValueError(
                "freeze operation is not available for StudioDependencyPolicy")

        needs_gpu = self._needs_gpu(resources_needed)
        packages = freeze.freeze()
        result = []
        for pkg in packages:
            if self._is_special_reference(pkg):
                # git repo or directly installed package
                result.append(pkg)
            elif '==' in pkg:
                # pypi package
                pkey = re.search(r'^.*?(?=\=\=)', pkg).group(0)
                pversion = re.search(r'(?<=\=\=).*\Z', pkg).group(0)

                if needs_gpu and \
                        (pkey == 'tensorflow' or pkey == 'tf-nightly'):
                    pkey = pkey + '-gpu'

                # TODO add installation logic for torch
                result.append(pkey + '==' + pversion)
        return result

    def _is_special_reference(self, pkg: str):
        if pkg.startswith('-e git+'):
            return True
        if 'git+https://' in pkg or 'file://' in pkg:
            return True
        return False

    def _needs_gpu(self, resources_needed):
        return resources_needed is not None and \
            int(resources_needed.get('gpus')) > 0




import os
import uuid
import time
from typing import Dict

from studio.storage.storage_type import StorageType
from studio.util.util import get_temp_filename, check_for_kb_interrupt

# StorageHandler encapsulates the logic of basic storage operations
# for specific storage endpoint (S3, http, local etc.)
# together with access credentials for this endpoint.
class StorageHandler:
    def __init__(self, storage_type: StorageType,
                 logger,
                 measure_timestamp_diff=False,
                 compression=None):
        self.type = storage_type
        self.logger = logger
        self.compression = compression
        self._timestamp_shift = 0
        if measure_timestamp_diff:
            try:
                self._timestamp_shift = self._measure_timestamp_diff()
            except BaseException:
                check_for_kb_interrupt()
                self._timestamp_shift = 0

    @classmethod
    def get_id(cls, config: Dict) -> str:
        raise NotImplementedError("Not implemented: upload_file")

    def upload_file(self, key, local_path):
        raise NotImplementedError("Not implemented: upload_file")

    def download_file(self, key, local_path):
        raise NotImplementedError("Not implemented: download_file")

    def download_remote_path(self, remote_path, local_path):
        raise NotImplementedError("Not implemented: download_remote_path")

    def delete_file(self, key, shallow=True):
        raise NotImplementedError("Not implemented: delete_file")

    def get_file_url(self, key, method='GET'):
        raise NotImplementedError("Not implemented: get_file_url")

    def get_file_timestamp(self, key):
        raise NotImplementedError("Not implemented: get_file_timestamp")

    def get_qualified_location(self, key):
        raise NotImplementedError("Not implemented: get_qualified_location")

    def get_local_destination(self, remote_path: str):
        raise NotImplementedError("Not implemented: get_local_destination")

    def get_timestamp_shift(self):
        return self._timestamp_shift

    def get_compression(self):
        return self.compression

    def cleanup(self):
        pass

    def _measure_timestamp_diff(self):
        max_diff = 60
        tmpfile = get_temp_filename() + '.txt'
        with open(tmpfile, 'w') as f:
            f.write('timestamp_diff_test')
        key = 'tests/' + str(uuid.uuid4())
        self.upload_file(key, tmpfile)
        remote_timestamp = self.get_file_timestamp(key)

        if remote_timestamp is not None:
            now_remote_diff = time.time() - remote_timestamp
            self.storage_handler.delete_file(key)
            os.remove(tmpfile)

            assert -max_diff < now_remote_diff and \
                now_remote_diff < max_diff, \
                "Timestamp difference is more than 60 seconds. " + \
                "You'll need to adjust local clock for caching " + \
                "to work correctly"

            return -now_remote_diff if now_remote_diff < 0 else 0




import sys
from Crypto.PublicKey import RSA
from Crypto.Cipher import PKCS1_OAEP
import nacl.secret
import nacl.utils
import base64

class Encryptor:
    """
    Implementation for experiment payload builder
    using public key RSA encryption.
    """
    def __init__(self, keypath: str):
        """
        param: keypath - file path to .pem file with public key
        """

        self.key_path = keypath
        self.recipient_key = None
        try:
            self.recipient_key = RSA.import_key(open(self.key_path).read())
        except:
            print(
                "FAILED to import recipient public key from: {0}".format(self.key_path))
            return

    def _import_rsa_key(self, key_path: str):
        key = None
        try:
            key = RSA.import_key(open(key_path).read())
        except:
            self.logger.error(
                "FAILED to import RSA key from: {0}".format(key_path))
            key = None
        return key

    def _encrypt_str(self, workload: str):
        # Generate one-time symmetric session key:
        session_key = nacl.utils.random(32)

        # Encrypt the data with the NaCL session key
        data_to_encrypt = workload.encode("utf-8")
        box_out = nacl.secret.SecretBox(session_key)
        encrypted_data = box_out.encrypt(data_to_encrypt)
        encrypted_data_text = base64.b64encode(encrypted_data)

        # Encrypt the session key with the public RSA key
        cipher_rsa = PKCS1_OAEP.new(self.recipient_key)
        encrypted_session_key = cipher_rsa.encrypt(session_key)
        encrypted_session_key_text = base64.b64encode(encrypted_session_key)

        return encrypted_session_key_text, encrypted_data_text

    def _decrypt_data(self, private_key_path, encrypted_key_text, encrypted_data_text):
        private_key = self._import_rsa_key(private_key_path)
        if private_key is None:
            return None

        try:
            private_key = RSA.import_key(open(private_key_path).read())
        except:
            self.logger.error(
                "FAILED to import private key from: {0}".format(private_key_path))
            return None

        # Decrypt the session key with the private RSA key
        cipher_rsa = PKCS1_OAEP.new(private_key)
        session_key = cipher_rsa.decrypt(
            base64.b64decode(encrypted_key_text))

        # Decrypt the data with the NaCL session key
        box_in = nacl.secret.SecretBox(session_key)
        decrypted_data = box_in.decrypt(
            base64.b64decode(encrypted_data_text))
        decrypted_data = decrypted_data.decode("utf-8")

        return decrypted_data

    def encrypt(self, payload: str):
        enc_key, enc_payload = self._encrypt_str(payload)

        enc_key_str = enc_key.decode("utf-8")
        enc_payload_str = enc_payload.decode("utf-8")

        return "{0},{1}".format(enc_key_str, enc_payload_str)

def main():
    if len(sys.argv) < 3:
        print("USAGE {0} public-key-file-path string-to-encrypt"
              .format(sys.argv[0]))
        return

    encryptor = Encryptor(sys.argv[1])
    data = sys.argv[2]
    print(data)
    result = encryptor.encrypt(data)
    print(result)

if __name__ == '__main__':
    main()



import time
from concurrent.futures import ThreadPoolExecutor, wait

from studio.util import util, logs
from studio.storage.storage_handler import StorageHandler
from studio.artifacts.artifact import Artifact
from studio.experiments.experiment import Experiment, experiment_from_dict
from studio.storage.storage_setup import get_storage_verbose_level
from studio.util.util import retry, report_fatal,\
    compression_to_extension, check_for_kb_interrupt


class KeyValueProvider:
    """Data provider for managing experiment lifecycle."""

    def __init__(
            self,
            db_config,
            handler: StorageHandler,
            compression=None):
        self.logger = logs.getLogger(self.__class__.__name__)
        self.logger.setLevel(get_storage_verbose_level())

        self.compression = compression
        if self.compression is None:
            self.compression = db_config.get('compression', None)

        self.auth = None

        self.storage_handler = handler

        self.max_keys = db_config.get('max_keys', 100)

    def _report_fatal(self, msg: str):
        report_fatal(msg, self.logger)

    def get_logger(self):
        return self.logger

    def _get_userid(self):
        userid = None
        if self.auth:
            userid = self.auth.get_user_id()
        userid = userid if userid else 'guest'
        return userid

    def _get_user_keybase(self, userid=None):
        if userid is None:
            userid = self._get_userid()

        return "users/" + userid + "/"

    def _get_experiments_keybase(self):
        return "experiments/"

    def _get_projects_keybase(self):
        return "projects/"

    def _experiment_key(self, experiment):
        if not isinstance(experiment, str):
            key = experiment.key
        else:
            key = experiment
        return key

    def add_experiment(self, experiment: Experiment,
                       userid=None, compression=None):
        self._delete(self._get_experiments_keybase() + experiment.key)
        experiment.time_added = time.time()
        experiment.status = 'waiting'

        compression = compression if compression else self.compression

        for tag, item in experiment.artifacts.items():
            art: Artifact = item
            if art.is_mutable:
                art.key = self._get_mutable_artifact_key(experiment, tag)
            else:
                if art.local_path is not None:
                    # upload immutable artifacts
                    art.key = art.upload(art.local_path)
                elif art.hash is not None:
                    art.key = self._get_immutable_artifact_key(
                        art.hash,
                        compression=compression
                    )

            if art.key is not None and art.remote_path is None:
                art.remote_path = art.storage_handler.get_qualified_location(art.key)

        userid = userid if userid else self._get_userid()
        experiment.owner = userid

        experiment_dict = experiment.to_dict()

        self._set(self._get_experiments_keybase() + experiment.key,
                  experiment_dict)

        if not experiment.from_compl_service:
            self._set(self._get_user_keybase(userid) + "experiments/" +
                  experiment.key,
                  experiment.time_added)

            if experiment.project and userid:
                self._set(self._get_projects_keybase() +
                      experiment.project + "/" +
                      experiment.key + "/owner",
                      userid)

        retry(lambda: self.checkpoint_experiment(experiment),
              sleep_time=10,
              logger=self.logger)
        self.logger.info("Added experiment %s", experiment.key)

    def _get_immutable_artifact_key(self, arthash, compression=None):
        retval = "blobstore/" + arthash + ".tar" + \
                 compression_to_extension(compression)
        return retval

    def _get_mutable_artifact_key(self, experiment: Experiment, tag: str) -> str:
        return self._get_experiments_keybase() + \
            experiment.key + '/' + tag + '.tar'

    def start_experiment(self, experiment):
        time_started = time.time()
        experiment.time_started = time_started
        experiment.status = 'running'

        experiment_dict = self._get(self._get_experiments_keybase() +
                                    experiment.key)
        experiment_dict['time_started'] = time_started
        experiment_dict['status'] = 'running'

        self._set(self._get_experiments_keybase() +
                  experiment.key,
                  experiment_dict)

        self.checkpoint_experiment(experiment)

    def stop_experiment(self, key):
        # can be called remotely (the assumption is
        # that remote worker checks experiments status periodically,
        # and if it is 'stopped', kills the experiment)
        key = self._experiment_key(key)

        experiment_data = self._get(self._get_experiments_keybase() +
                                    key)

        experiment_data['status'] = 'stopped'

        self._set(self._get_experiments_keybase() +
                  key, experiment_data)

    def finish_experiment(self, experiment):
        time_finished = time.time()
        key = self._experiment_key(experiment)
        if not isinstance(experiment, str):
            experiment.status = 'finished'
            experiment.time_finished = time_finished

        experiment_dict = self._get(
            self._get_experiments_keybase() + key)

        experiment_dict['status'] = 'finished'
        experiment_dict['time_finished'] = time_finished

        self._set(self._get_experiments_keybase() +
                  key, experiment_dict)

    def delete_experiment(self, experiment):
        if isinstance(experiment, str):
            experiment_key = experiment
            try:
                experiment = self.get_experiment(experiment)
                experiment_key = experiment.key
            except BaseException:
                check_for_kb_interrupt()
                experiment = None
        else:
            experiment_key = experiment.key

        experiment_dict = self._get(self._get_experiments_keybase() +
                                    experiment_key)
        if experiment_dict is None:
            self.logger.error("FAILED to delete experiment %s: NOT FOUND.",
                              experiment_key)
            return

        from_compl_service: bool = experiment_dict.get('from_compl_service', False)
        experiment_owner = experiment_dict.get('owner')
        if experiment_owner is not None and not from_compl_service:
            self._delete(self._get_user_keybase(experiment_owner) +
                     'experiments/' + experiment_key, shallow=False)

        if experiment is not None:
            for tag, art in experiment.artifacts.items():
                if art.key is not None and art.is_mutable:
                    msg: str =\
                        ('Deleting artifact {0} from the store, ' +
                         'artifact key {1}').format(tag, art.key)
                    self.logger.debug(msg)
                    art.delete()

            if experiment.project is not None and not from_compl_service:
                self._delete(
                    self._get_projects_keybase() +
                    experiment.project + "/" + experiment_key + "/" + "owner",
                    shallow=False)

        self._delete(self._get_experiments_keybase() + experiment_key, shallow=False)


    def cleanup(self):
        if self.storage_handler is not None:
            self.storage_handler.cleanup()


    def checkpoint_experiment(self, experiment):
        key = self._experiment_key(experiment)
        key_path = self._get_experiments_keybase() + key

        self.logger.debug('checkpointing %s', key_path)

        experiment_dict = self._get(key_path)

        workers = []
        with ThreadPoolExecutor(max_workers=8) as executor:
            for _, art in experiment.artifacts.items():
                if art.is_mutable and art.local_path is not None:
                    workers.append(executor.submit(art.upload, None))
            wait(workers)

        for worker in workers:
            try:
                worker.result()
            except Exception as exc:
                # If any of artifact savers failed and threw an exception,
                # rethrow it to signal overall checkpoint failure
                raise exc

        checkpoint_time = time.time()
        experiment.time_last_checkpoint = checkpoint_time
        experiment_dict['time_last_checkpoint'] = checkpoint_time

        self._set(key_path, experiment_dict)

    def _get_experiment_info(self, experiment: Experiment):
        info = {}
        type_found = False

        if not type_found:
            info['type'] = 'unknown'

        info['logtail'] = self._get_experiment_logtail(experiment)
        return info

    def _get_experiment_logtail(self, experiment: Experiment):
        try:
            tarf = experiment.artifacts['output'].stream()
            if not tarf:
                return None
            tarf_member = tarf.members[0]
            while tarf_member is not None and tarf_member.name == '':
                tarf_member = tarf.next()

            logdata = tarf.extractfile(tarf_member).read()
            logdata = util.remove_backspaces(logdata).split('\n')
            return logdata
        except BaseException as exc:
            self.logger.debug('Getting experiment logtail raised an exception: %s',
                              repr(exc))
            check_for_kb_interrupt()
            return None

    def get_experiment(self, key, getinfo=True) -> Experiment:
        data = self._get(self._get_experiments_keybase() + key)
        if data is None:
            return None

        data['key'] = key

        experiment_stub = experiment_from_dict(data)

        expinfo = {}
        if getinfo:
            try:
                expinfo = self._get_experiment_info(experiment_stub)
            except Exception as exc:
                self.logger.info(
                    "Exception %s while info download for %s", repr(exc), key)

        return experiment_from_dict(data, expinfo)

    def get_user_experiments(self, userid=None):
        if userid and '@' in userid:
            users = self.get_users()
            user_ids = [u for u in users if users[u].get('email') == userid]
            if len(user_ids) < 1:
                return None
            userid = user_ids[0]

        experiment_keys = self._get(
            self._get_user_keybase(userid) + "experiments/", shallow=True)
        if not experiment_keys:
            experiment_keys = []

        return experiment_keys

    def get_project_experiments(self, project):
        experiment_keys = self._get(self._get_projects_keybase() +
                                    project)
        if not experiment_keys:
            experiment_keys = {}

        return experiment_keys.keys()

    def get_artifacts(self, key):
        if isinstance(key, str):
            experiment = self.get_experiment(key, getinfo=False)
        else:
            experiment = key

        retval = {}
        for tag, art in experiment.artifacts.items():
            url = art.get_url()
            if url is not None:
                retval[tag] = url

        return retval

    def get_artifact(self, artifact, local_path=None, only_newer=True):
        return artifact.download(
            local_path=local_path,
            only_newer=only_newer)

    def get_projects(self):
        return self._get(self._get_projects_keybase(), shallow=True)

    def get_users(self):
        user_ids = self._get('users/', shallow=True)
        retval = {}
        if user_ids:
            for user_id in user_ids:
                retval[user_id] = {
                    'email': self._get('users/' + user_id + '/email')
                }
        return retval

    def refresh_auth_token(self, email, refresh_token):
        if self.auth:
            self.auth.refresh_token(email, refresh_token)

    def is_auth_expired(self):
        if self.auth:
            return self.auth.expired
        return False

    def can_write_experiment(self, key=None, user=None):
        assert key is not None
        user = user if user else self._get_userid()

        experiment = self._get(
            self._get_experiments_keybase() + key)

        if experiment:
            owner = experiment.get('owner')
            if owner is None or owner == 'guest':
                return True
            return owner == user
        return True

    def register_user(self, userid, email):
        keypath = self._get_user_keybase(userid) + 'email'
        existing_email = self._get(keypath)
        if existing_email != email:
            self._set(keypath, email)

    def get_storage_handler(self):
        return self.storage_handler

    def _get(self, key, shallow=False):
        raise NotImplementedError("Not implemented: _get")

    def _set(self, key, value):
        raise NotImplementedError("Not implemented: _set")

    def _delete(self, key, shallow=True):
        raise NotImplementedError("Not implemented: _delete")

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass



import threading

from studio.queues.rabbit_queue import RMQueue
from studio.util.util import check_for_kb_interrupt

_queue_cache = {}

def get_cached_queue(
        name,
        route,
        cloud=None,
        config=None,
        logger=None,
        close_after=None,
        verbose=10):

    q = _queue_cache.get(name, None)
    if q is not None:
        if logger is not None:
            logger.info("Got queue named {0} from queue cache."
                        .format(name))
        return q

    q = RMQueue(
        queue=name,
        route=route,
        config=config,
        logger=logger)

    if logger is not None:
        logger.info("Created new queue named {0}.".format(name))

    if close_after is not None and close_after.total_seconds() > 0:
        thr = threading.Timer(
            interval=close_after.total_seconds(),
            function=purge_rmq,
            kwargs={
                "q": q,
                "logger": logger})
        thr.setDaemon(True)
        thr.start()

    _queue_cache[name] = q
    if logger is not None:
        logger.info("Added queue named {0} to queue cache."
                    .format(name))
    return q

def shutdown_cached_queue(queue, logger=None, delete_queue=True):
    if queue is None:
        return

    _queue_cache.pop(queue.get_name(), None)
    if logger is not None:
        logger.info("Removed queue named {0} from queue cache."
                    .format(queue.get_name()))

    queue.shutdown(delete_queue)


def purge_rmq(q, logger, **kwargs):
    if q is None:
        return

    try:
        q.shutdown(True)
    except BaseException as e:
        check_for_kb_interrupt()
        logger.warning(e)
        return
    return



import os
import shutil
from typing import Dict

from studio.storage.storage_setup import get_storage_verbose_level
from studio.storage.storage_type import StorageType
from studio.storage.storage_handler import StorageHandler
from studio.util import logs
from studio.util import util

class LocalStorageHandler(StorageHandler):
    def __init__(self, config,
                 measure_timestamp_diff=False,
                 compression=None):

        self.logger = logs.getLogger(self.__class__.__name__)
        self.logger.setLevel(get_storage_verbose_level())

        if compression is None:
            compression = config.get('compression')

        self.endpoint = config.get('endpoint', '~')
        self.endpoint = os.path.realpath(os.path.expanduser(self.endpoint))
        if not os.path.exists(self.endpoint) \
            or not os.path.isdir(self.endpoint):
            msg: str = "Store root {0} doesn't exist or not a directory. Aborting."\
                .format(self.endpoint)
            self._report_fatal(msg)

        self.bucket = config.get('bucket', 'storage')
        self.store_root = os.path.join(self.endpoint, self.bucket)
        self._ensure_path_dirs_exist(self.store_root)

        super().__init__(StorageType.storageLocal,
            self.logger,
            measure_timestamp_diff,
            compression=compression)

    def _ensure_path_dirs_exist(self, path):
        dirs = os.path.dirname(path)
        os.makedirs(dirs, mode = 0o777, exist_ok = True)

    def _copy_file(self, from_path, to_path):
        try:
            shutil.copyfile(from_path, to_path)
        except Exception as exc:
            msg: str = "FAILED to copy '{0}' to '{1}': {2}. Aborting."\
                .format(from_path, to_path,exc)
            self._report_fatal(msg)

    def upload_file(self, key, local_path):
        target_path = os.path.join(self.store_root, key)
        if not os.path.exists(local_path):
            self.logger.debug(
                "Local path {0} does not exist. SKIPPING upload to {1}"
                    .format(local_path, target_path))
            return False
        self._ensure_path_dirs_exist(target_path)
        self._copy_file(local_path, target_path)
        return True

    def download_file(self, key, local_path):
        source_path = os.path.join(self.store_root, key)
        if not os.path.exists(source_path):
            self.logger.debug(
                "Source path {0} does not exist. SKIPPING download to {1}"
                    .format(source_path, local_path))
            return False
        self._ensure_path_dirs_exist(local_path)
        self._copy_file(source_path, local_path)
        return True

    def delete_file(self, key, shallow=True):
        key_path: str = self._get_file_path_from_key(key)
        if os.path.exists(key_path):
            self.logger.debug("Deleting local file {0}.".format(key_path))
            util.delete_local_path(key_path, self.store_root, False)

    @classmethod
    def get_id(cls, config: Dict) -> str:
        endpoint = config.get('endpoint', None)
        if endpoint is None:
            return None
        return '[local]{0}'.format(endpoint)

    def _get_file_path_from_key(self, key: str):
        return str(os.path.join(self.store_root, key))

    def get_file_url(self, key, method='GET'):
        return self._get_file_path_from_key(key)

    def get_file_timestamp(self, key):
        key_path: str = self._get_file_path_from_key(key)
        if os.path.exists(key_path):
            return os.path.getmtime(key_path)
        else:
            return None

    def get_qualified_location(self, key):
        return 'file:/' + self.store_root + '/' + key

    def get_endpoint(self):
        return self.endpoint

    def _report_fatal(self, msg: str):
        util.report_fatal(msg, self.logger)



import os
from urllib.parse import urlparse
from typing import Dict
from studio.util import logs
from studio.credentials.credentials import Credentials
from studio.storage.storage_setup import get_storage_verbose_level
from studio.storage.storage_type import StorageType
from studio.storage.storage_handler import StorageHandler
from studio.storage import storage_util

class HTTPStorageHandler(StorageHandler):
    def __init__(self, remote_path, credentials_dict,
                 timestamp=None,
                 compression=None):

        self.logger = logs.getLogger(self.__class__.__name__)
        self.logger.setLevel(get_storage_verbose_level())

        self.url = remote_path
        self.timestamp = timestamp

        parsed_url = urlparse(self.url)
        self.scheme = parsed_url.scheme
        self.endpoint = parsed_url.netloc
        self.path = parsed_url.path
        self.credentials = Credentials(credentials_dict)

        super().__init__(StorageType.storageHTTP,
            self.logger,
            False,
            compression=compression)

    def upload_file(self, key, local_path):
        storage_util.upload_file(self.url, local_path, self.logger)

    def download_file(self, key, local_path):
        return storage_util.download_file(self.url, local_path, self.logger)

    def download_remote_path(self, remote_path, local_path):
        head, _ = os.path.split(local_path)
        if head is not None:
            os.makedirs(head, exist_ok=True)
        return storage_util.download_file(remote_path, local_path, self.logger)

    @classmethod
    def get_id(cls, config: Dict) -> str:
        endpoint = config.get('endpoint', None)
        if endpoint is None:
            return None
        creds: Credentials = Credentials.getCredentials(config)
        creds_fingerprint = creds.get_fingerprint() if creds else ''
        return '[http]{0}::{1}'.format(endpoint, creds_fingerprint)

    def get_local_destination(self, remote_path: str):
        parsed_url = urlparse(remote_path)
        parts = parsed_url.path.split('/')
        return None, parts[len(parts)-1]

    def delete_file(self, key, shallow=True):
        raise NotImplementedError

    def get_file_url(self, key):
        return self.url

    def get_file_timestamp(self, key):
        return self.timestamp

    def get_qualified_location(self, key):
        return self.url



import hashlib
from typing import Dict

from studio.util import logs, util
from studio.storage import storage_setup

AWS_KEY = 'access_key'
AWS_SECRET_KEY = 'secret_access_key'
AWS_SESSION_TOKEN = 'session_token'
AWS_REGION = 'region'
AWS_PROFILE = 'profile'
KEY_CREDENTIALS = 'credentials'
AWS_TYPE = 'aws'

# Credentials encapsulates the logic
# of basic access credentials handling
# for different kinds of studioml storage providers (S3, http, local etc.)
class Credentials:
    def __init__(self, cred_dict):
        self.logger = logs.getLogger(self.__class__.__name__)
        self.logger.setLevel(storage_setup.get_storage_verbose_level())

        self.type = None
        self.key = None
        self.secret_key = None
        self.session_token = None
        self.region = None
        self.profile = None
        if cred_dict is None:
            return

        if isinstance(cred_dict, str) and cred_dict == 'none':
            return

        if not isinstance(cred_dict, dict):
            msg: str =\
                "NOT SUPPORTED credentials format {0}".format(repr(cred_dict))
            util.report_fatal(msg, self.logger)

        if len(cred_dict) == 0:
            # Empty credentials dictionary is like None:
            return

        if len(cred_dict) == 1 and AWS_TYPE in cred_dict.keys():
            aws_creds = cred_dict[AWS_TYPE]
            self.type = AWS_TYPE
            self.key = aws_creds.get(AWS_KEY, None)
            self.secret_key = aws_creds.get(AWS_SECRET_KEY, None)
            self.session_token = aws_creds.get(AWS_SESSION_TOKEN, None)
            self.region = self._get_named(AWS_REGION, aws_creds)
            self.profile = self._get_named(AWS_PROFILE, aws_creds)

            if self.key is None or self.secret_key is None:
                msg: str = \
                    "INVALID aws credentials format {0}".format(repr(cred_dict))
                util.report_fatal(msg, self.logger)
        else:
            msg: str =\
                "NOT SUPPORTED credentials format {0}".format(repr(cred_dict))
            util.report_fatal(msg, self.logger)

    def get_type(self):
        return self.type

    def get_key(self):
        return self.key

    def get_secret_key(self):
        return self.secret_key

    def get_session_token(self):
        return self.session_token

    def get_region(self):
        return self.region

    def get_profile(self):
        return self.profile

    def _get_named(self, key: str, config: Dict):
        value = config.get(key, None)
        if value is None:
            value = config.get(key+'_name', None)
        return value

    def to_dict(self):
        result: Dict = {
            self.type: {
                AWS_KEY: self.key,
                AWS_SECRET_KEY: self.secret_key
            }
        }
        if self.session_token is not None:
            result[self.type][AWS_SESSION_TOKEN] = self.session_token
        if self.region is not None:
            result[self.type][AWS_REGION] = self.region
        if self.profile is not None:
            result[self.type][AWS_PROFILE] = self.profile
        return result

    def get_fingerprint(self) -> str:
        if self.type is None and self.key is None and\
            self.secret_key is None:
            return ''
        id: str = "{0}::{1}::{2}::{3}::{4}::{5}"\
            .format(self.type, self.key, self.secret_key,
                    self.session_token,
                    self.profile, self.region)
        return hashlib.sha256(id.encode()).hexdigest()

    @classmethod
    def getCredentials(cls, config):
        if config is None:
            return None
        cred_dict = config.get(KEY_CREDENTIALS, None)
        return Credentials(cred_dict) if cred_dict else None

    def to_string(self):
        return "type: {0} key: {1} secret: <reducted> profile: {2} region: {3}"\
            .format(self.type, self.key, self.profile, self.region)

